/*
 *	LCD, LED, and Button interface for Cobalt	
 *      Andrew Bose
 *	Timothy Stonis (x86 version)
 * 	Tim Hockin
 * 	Adrian Sun
 *
 */
#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/mc146818rtc.h>
#include <linux/netdevice.h>
#include <linux/proc_fs.h>
#include <linux/in6.h>
#include <linux/pci.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/checksum.h>
#include <linux/delay.h>
#include <linux/lcd.h>

#define LCD_DRIVER	"lcd: Cobalt LCD Driver v3.01 by\r\nlcd: Andrew Bose <andrew@cobaltnet.com>, Timothy Stonis <tim@cobaltnet.com>"

/* io registers */
#define LPT                   0x0378
#define LCD_DATA_ADDRESS      LPT+0
#define LCD_CONTROL_ADDRESS   LPT+2

#define kLCD_Addr             0x00000080
#define kDD_R00         0x00
#define kDD_R01         0x27
#define kDD_R10         0x40
#define kDD_R11         0x67

/* led bits */
#define LED_SHUTCOBALT_ADDR   0x7e
#define LED_SHUTDOWN_OFF      0x40
#define LED_COBALTLOGO_OFF    0x80

#define LED_WEBLIGHT_ADDR     0xb8
#define LED_WEBLIGHT_ON       0x80


static int lcd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
                     unsigned long arg);

static struct pci_dev *led_dev = NULL;
static unsigned char led_state = 0;

static int lcd_present = 1;
static struct proc_dir_entry *proc_lcd;

/* 
 * LEDSet: set the lcds
 */
static inline void LEDSet(const unsigned char ledstate)
{
       __u8 tmp;

       if (!led_dev)
	 return;

       /* handle shutdown and cobalt logo led at the same time
	* note: the cobalt logo has reverse polarity */
       pci_read_config_byte(led_dev, LED_SHUTCOBALT_ADDR, &tmp);
       if (ledstate & LED_SHUTDOWN) {
	       tmp &= ~LED_SHUTDOWN_OFF;
       } else {
	       tmp |= LED_SHUTDOWN_OFF;
       }

       if (ledstate & LED_COBALTLOGO) {
	       tmp &= ~LED_COBALTLOGO_OFF;
       } else {
	       tmp |= LED_COBALTLOGO_OFF;
       }
       pci_write_config_byte(led_dev, LED_SHUTCOBALT_ADDR, tmp);

       /* web light */
       pci_read_config_byte(led_dev, LED_WEBLIGHT_ADDR, &tmp);
       if (ledstate & LED_WEBLIGHT) {
	       tmp |= LED_WEBLIGHT_ON;
       } else {
	       tmp &= ~LED_WEBLIGHT_ON;
       }
       pci_write_config_byte(led_dev, LED_WEBLIGHT_ADDR, tmp);
}

/*
** LCDReadInst: Read a control instruction from the LCD
*/
static inline unsigned char LCDReadInst (void)
{
        unsigned char a;
        outb(0x21,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0
        outb(0x20,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=1
        a=inb(LCD_DATA_ADDRESS);
        outb(0x21,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0
        outb(0x01,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0

        // This is bad... use jiffies.
        udelay(100);                       // 10ms delay

        return a;
}

/*
** LCDWriteInst: Write a control instruction to the LCD
*/
static inline void LCDWriteInst (unsigned char data)
{
	while ((LCDReadInst() & 0x80) == 0x80);

        outb(0x03,LCD_CONTROL_ADDRESS);      // RS=0, R/W=0, E=0
        outb(data,LCD_DATA_ADDRESS);
        outb(0x02,LCD_CONTROL_ADDRESS);      // RS=0, R/W=0, E=1
        outb(0x03,LCD_CONTROL_ADDRESS);      // RS=0, R/W=0, E=0
        outb(0x01,LCD_CONTROL_ADDRESS);      // RS=0, R/W=1, E=0

        // This is bad... use jiffies.
        udelay(100);                       // 10ms delay
}


/*
** LCDWriteData: Write one byte of data to the LCD
*/
static inline void LCDWriteData (unsigned char data)
{
	while ((LCDReadInst() & 0x80) == 0x80);

        outb(0x07,LCD_CONTROL_ADDRESS);      // RS=1, R/W=0, E=0
        outb(data,LCD_DATA_ADDRESS);
        outb(0x06,LCD_CONTROL_ADDRESS);      // RS=1, R/W=0, E=1
        outb(0x07,LCD_CONTROL_ADDRESS);      // RS=1, R/W=0, E=0
        outb(0x05,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0

        // This is bad... use jiffies.
        udelay(100);                         // 1ms delay
}

/*
** LCDReadData: Read one byte of data from the LCD
*/
static inline unsigned char LCDReadData (void)
{
        unsigned char a;

	while ((LCDReadInst() & 0x80) == 0x80);

        outb(0x25,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0
        outb(0x24,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=1
        a = inb(LCD_DATA_ADDRESS);
        outb(0x25,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0
        outb(0x01,LCD_CONTROL_ADDRESS);      // RS=1, R/W=1, E=0

        // This is bad... use jiffies.
        udelay(100);                        // 1ms delay

        return a;
}

static inline void LCD_InitDriver (void)
{
        LCDWriteInst(0x38);
        LCDWriteInst(0x38);
        LCDWriteInst(0x38);
        LCDWriteInst(0x06);
        LCDWriteInst(0x0c);
}

static inline char ReadButtons ( void )
{
  outb(0x29,LCD_CONTROL_ADDRESS);  // Sel=0, Bi=1
  return( inb(LCD_DATA_ADDRESS) );
}

static inline int button_pressed(void)
{
        unsigned char buttons = ReadButtons();

        if ( (buttons == BUTTON_Next) || (buttons == BUTTON_Next_B) ||
             (buttons == BUTTON_Reset_B) )
                return buttons;
        return 0;
}


static int lcd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
			unsigned long arg)
{
	struct lcd_display button_display;
	unsigned long address, a;
	int index;

	switch (cmd) {
	/* Turn the LCD on */
	case LCD_On:
		LCDWriteInst(0x0F);
		break;		
	/* Turn the LCD off */
	case LCD_Off:
		LCDWriteInst(0x08);
		break;
	/* Reset the LCD */
	case LCD_Reset:
		LCDWriteInst(0x3F);
		LCDWriteInst(0x3F);
		LCDWriteInst(0x3F);
		LCDWriteInst(0x3F);
		LCDWriteInst(0x01);
		LCDWriteInst(0x06);
		break;

	/* Clear the LCD */
	case LCD_Clear:
       		LCDWriteInst(0x01);     
		break;

	/* Move the cursor one position to the left */
	case LCD_Cursor_Left:
		LCDWriteInst(0x10);
		break;

	/* Move the cursor one position to the right */
	case LCD_Cursor_Right:
		LCDWriteInst(0x14);
		break;	

	/* Turn the cursor off */
	case LCD_Cursor_Off:
                LCDWriteInst(0x0C);
	        break;

	/* Turn the cursor on */
        case LCD_Cursor_On:
                LCDWriteInst(0x0F);
                break;

	/* Turn blinking off? I don't know what this does - TJS */
        case LCD_Blink_Off:
                LCDWriteInst(0x0E);
                break;

	/* Get the current cursor position */
        case LCD_Get_Cursor_Pos:{
                struct lcd_display display;

                display.cursor_address = ( LCDReadInst() );
                display.cursor_address = ( display.cursor_address & 0x07F );
                copy_to_user((struct lcd_display*)arg, &display, 
		             sizeof(struct lcd_display));

                break;
                }

	/* Set the cursor position */
        case LCD_Set_Cursor_Pos: {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg, 
                              sizeof(struct lcd_display));
                a = (display.cursor_address | kLCD_Addr );
                LCDWriteInst( a );
                break;
                }
	/* Get the value at the current cursor position? - TJS */
        case LCD_Get_Cursor: {
                struct lcd_display display;
                display.character = LCDReadData();
                copy_to_user((struct lcd_display*)arg, &display, 
		             sizeof(struct lcd_display));
                LCDWriteInst(0x10);
                break;
                }

	/* Set the character at the cursor position? - TJS */
	case LCD_Set_Cursor:{
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg, 
                              sizeof(struct lcd_display));
                LCDWriteData( display.character );
                LCDWriteInst(0x10);
                break;
                }

	/* Dunno what this does - TJS */ 
	case LCD_Disp_Left:
		LCDWriteInst(0x18);
		break;

	/* Dunno what this does - TJS */ 
	case LCD_Disp_Right:
		LCDWriteInst(0x1C);
		break;

	/* Dunno what this does - TJS */ 
	case LCD_Home:
		LCDWriteInst(0x02);
		break;

	/* Write a string to the LCD */
	case LCD_Write: {
		struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg, 
                              sizeof(struct lcd_display));
                LCDWriteInst(0x80);

		/* First line */	
		for (index = 0; index < (display.size1); index++) {
			LCDWriteData( display.line1[index]);
		}		
		LCDWriteInst(0xC0);	

		/* Second line */
                for (index = 0; index < (display.size2); index++) {
                        LCDWriteData( display.line2[index]);
		}
 
		break;	
	}

	/* Read what's on the LCD */
        case LCD_Read: {
                struct lcd_display display;

                for (address = kDD_R00; address <= kDD_R01; address++) {
                        a = (address | kLCD_Addr );

                        LCDWriteInst( a );
                        display.line1[address] = LCDReadData();
                }
                display.line1[ 0x27 ] = '\0';
                for (address = kDD_R10; address <= kDD_R11; address++) {
                        a = (address | kLCD_Addr );
                        LCDWriteInst( a );
                        display.line2[address - 0x40 ] = LCDReadData();
                 }
                display.line2[ 0x27 ] = '\0';
                copy_to_user((struct lcd_display*)arg, &display,
                            sizeof(struct lcd_display));
                break;
        }

	case LCD_Raw_Inst: {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
                LCDWriteInst( display.character );
                break;
                }

        case LCD_Raw_Data: {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
                LCDWriteData( display.character );
                break;
                }

	/* Read the buttons */
        case BUTTON_Read: {
                button_display.buttons = ReadButtons();;
                copy_to_user((struct lcd_display*)arg, &button_display, 
		             sizeof(struct lcd_display));
                break;
        }
	
	/* set all of the leds */
	case LED_Set:
	  {
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
		led_state = display.leds;
                LEDSet(led_state);
		break;
	  }

	/* set a single led */
	case LED_Bit_Set:
	  { 
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
		led_state |= display.leds;
                LEDSet( led_state );
	  }
	  break;

	/* clear an led */
	case LED_Bit_Clear:
	  { 
                struct lcd_display display;
                copy_from_user(&display, (struct lcd_display*)arg,
                              sizeof(struct lcd_display));
		led_state &= ~display.leds;
                LEDSet( led_state );
	  }
	  break;

	default:
		return 0;
		break;

	}

	return 0;
}

static int lcd_open(struct inode *inode, struct file *file)
{
	if (!lcd_present)
		return -ENXIO;
	else
		return 0;
}

/* LED daemon sits on this and we wake him up once a key is pressed. */
static int lcd_waiters = 0;

static ssize_t lcd_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
        int buttons_now;

        if(lcd_waiters > 0)
                return -EINVAL;

        lcd_waiters++;
        while(((buttons_now = (int)button_pressed()) == 0) &&
              !(signal_pending(current))) {
                current->state = TASK_INTERRUPTIBLE;
                schedule_timeout((2 * HZ));
        }
        lcd_waiters--;

        if(signal_pending(current))
                return -ERESTARTSYS;
        return buttons_now;
}

/*
 *	The various file operations we support.
 */

static struct file_operations lcd_fops = {
	NULL,		/* No seek */
	lcd_read,
	NULL,		/* No write */ 
	NULL,		/* No readdir */
	NULL,    	/* No poll */
	lcd_ioctl,
	NULL,		/* No mmap */
	lcd_open,
	NULL		/* No release */	
};

/* read a single line from the LCD into a string */
char *LCDReadLine(int lineno)
{
	char *line;
	unsigned long address;
	unsigned long min, max;

	line = kmalloc(LCD_CHARS_PER_LINE, GFP_KERNEL);
	if (!line)
		return NULL;

	if (lineno == 0) {
		min = kDD_R00;
		max = kDD_R01;
	} else if (lineno == 1) {
		min = kDD_R10;
		max = kDD_R11;
	} else {
		min = 1;
		max = 0;
	}

	for (address = min; address <= max; address++) {
                unsigned long a = (address | kLCD_Addr );

                udelay(150);
                LCDWriteInst(a);

                udelay(150);
                line[address-min] = LCDReadData();
	}
	line[address-min] = '\0';

	return line;
}

/* write a string to a given line of the LCD */
int LCDWriteLine(int lineno, char *text)
{
	int i;
	char *blank = "                ";

	udelay(150);
	if (lineno == 0) {
		LCDWriteInst(0x80);
	} else if (lineno == 1) {
		LCDWriteInst(0xC0);
	}
	udelay(150);

	if (!text)
		text = blank;

	for (i = 0; i < 16; i++) {
		udelay(150);
		if (i < strlen(text)) {
			LCDWriteData(text[i]);
		} else {
			LCDWriteData(' ');
		}
	}

	return strlen(text);
}

int lcd_read_proc(char *buf, char **start, off_t offset, int len, int *eof, 
		  void *private)
{
	int plen = 0;
	char *tmp;

	tmp = LCDReadLine(0);
	if (tmp){
		plen += sprintf(buf+plen, "%s\n", tmp);
		kfree(tmp);
	} else {
		plen += sprintf(buf+plen, "\n");
	}
	tmp = LCDReadLine(1);
	if (tmp) {
		plen += sprintf(buf+plen, "%s\n", tmp);
		kfree(tmp);
	} else {
		plen += sprintf(buf+plen, "\n");
	}

        return plen;
}

static struct miscdevice lcd_dev=
{
	COBALT_LCD_MINOR,
	"lcd",
	&lcd_fops
};

void lcd_led_set(const unsigned char leds)
{
	led_state = leds;
	LEDSet(leds);
}

unsigned char lcd_led_get(void)
{
	return led_state;
}

int lcd_init(void)
{
	printk("%s\n", LCD_DRIVER);
	misc_register(&lcd_dev);

#ifdef CONFIG_PCI
	/* scan the pci bus for led ports */
	if (!pci_present())
	  return -1;

	led_dev = pci_find_device(PCI_VENDOR_ID_AL, PCI_DEVICE_ID_AL_M7101,
				  led_dev);
	if (!led_dev)
	  return -1;
	lcd_led_set(LED_SHUTDOWN);
#endif

        lcd_present = 1;
        LCD_InitDriver();
	proc_lcd = create_proc_entry("lcd", 0, NULL);
	if (proc_lcd) {
		proc_lcd->read_proc = lcd_read_proc;
	}

	return 0;
}
